local is_debug = false
local math_min = math.min
local table_insert = table.insert
local table_remove = table.remove

-- Raise the event of given name and parameters.
local function raise_event(event_name, param)
	param.name = event_name
	param.tick = game.tick
	param.mod = "copy-and-paste-modules"
	script.raise_event(event_name, param)
end

-- Creates an item of the given name at the given position and orders robot to take it away.
-- Returns whether the item can be created.
local function create_item_for_robot(player_index, surface, position, force, item_name)
	-- Spill onto the ground. LuaSurface.spill_item_stack does not return the item entity, so we have to use create_entity.
	local item_entity = surface.create_entity
	{
		name = "item-on-ground",
		position = position,
		stack = {name = item_name},
		fast_replace = true,
		spill = true
	}
	if item_entity then
		-- Item is created, order deconstruction.
		if item_entity.order_deconstruction(force) then
			-- Raise event.
			raise_event(defines.events.on_marked_for_deconstruction,
			{
				entity = item_entity,
				player_index = player_index
			})
		end
		return true
	end
	return false
end

-- Shows a flying text about item insertion or removal.
-- Returns the text position for the next flying text if the flying text is successfully created.
local function show_flying_text(player, surface, text_position, item_name, item_count)
	if surface.create_entity
	{
		name = "flying-text",
		position = text_position,
		text = {"flying-text.copy-and-paste-modules_inserted-items", item_count >= 0 and "+" or "", item_count, game.item_prototypes[item_name].localised_name, player.get_item_count(item_name)}
	} then
		text_position.y = text_position.y + 0.5
	end
	return text_position
end

-- Shows the flying texts about item insertion or removal according to the given dictionary of item name and count.
-- Returns the final text position for the next flying text.
local function show_flying_texts_for_all_items(player, surface, text_position, item_names_counts_dictionary)
	for name, count in pairs(item_names_counts_dictionary) do
		text_position = show_flying_text(player, surface, text_position, name, count)
	end
	return text_position
end

-- Removes all module requests in the given item request proxy.
local function remove_all_module_requests_in_item_request_proxy(item_request_proxy)
	local destroy_proxy = true
	local proxy_item_requests = item_request_proxy.item_requests
	for name, count in pairs(proxy_item_requests) do
		if game.item_prototypes[name].type == "module" then
			proxy_item_requests[name] = nil
		else
			-- The request is not a module. Keep the proxy.
			destroy_proxy = false
		end
	end
	if destroy_proxy then
		item_request_proxy.destroy()
	else
		item_request_proxy.item_requests = proxy_item_requests
	end
end

-- Removes all module requests in all of the given item request proxies.
local function remove_all_module_requests_in_item_request_proxies(item_request_proxies)
	for _, proxy in ipairs(item_request_proxies) do
		remove_all_module_requests_in_item_request_proxy(proxy)
	end
end

-- Removes all module requests on the given entity.
local function remove_all_module_requests_on_entity(entity, surface, position, force)
	local x = position.x
	local y = position.y
	local item_request_proxies = surface.find_entities_filtered
	{
		area = {{x - 0.1, y - 0.1}, {x + 0.1, y + 0.1}},
		name = "item-request-proxy",
		force = force
	}
	for _, proxy in ipairs(item_request_proxies) do
		if proxy.valid and proxy.proxy_target == entity then
			remove_all_module_requests_in_item_request_proxy(proxy)
		end
	end
end

-- Updates all module requests in the given dictionary of item requests mapped to proxies according to the given dictionary of item request count mapped with item name.
-- Also ensure that there are enough slots for the requests.
-- This method can works properly only if the new item requests dictionary has the same or fewer requests than the existing requests in the proxies.
local function update_item_requests_in_proxies(item_requests_by_proxy, new_item_requests_count_by_name, available_slots)
	if is_debug then game.print("Update item requests in proxies. Available slots: " .. available_slots) end
	for proxy, item_requests in pairs(item_requests_by_proxy) do
		for name, count in pairs(item_requests) do
			if available_slots > 0 then
				if new_item_requests_count_by_name[name] then
					count = math_min(new_item_requests_count_by_name[name], count, available_slots)
					item_requests[name] = count
					available_slots = available_slots - count
					count = new_item_requests_count_by_name[name] - count
					if count > 0 then
						-- We can keep more requests of the item.
						new_item_requests_count_by_name[name] = count
					else
						-- We don't need this item anymore.
						new_item_requests_count_by_name[name] = nil
					end
				else
					-- There are already enough requests for this item. Only remove module requests.
					if game.item_prototypes[name].type == "module" then
						item_requests[name] = nil
					end
				end
			else
				-- We don't have anymore slots. Only remove module requests.
				if game.item_prototypes[name].type == "module" then
					-- There are already enough requests for this module.
					item_requests[name] = nil
				end
			end
		end
		if next(item_requests) == nil then
			-- No more request for this proxy.
			item_requests_by_proxy[proxy] = nil
			proxy.destroy()
		else
			item_requests_by_proxy[proxy] = item_requests
			proxy.item_requests = item_requests
		end
	end
end

-- Removes all modules from the given entity. The removed modules will be inserted to player first. The exceeded ones may be taken away by robots, according to the given setting.
-- Module requests will also be removed.
-- Returns the text position for the next flying text.
-- @param text_position	Optional.
local function remove_all_modules_from_entity(entity, entity_module_inventory, player, player_index, surface, position, force, robot_removes_modules_after_player, text_position)
	local text_position = text_position or {x = position.x, y = position.y}
	-- Insert the modules to player's inventory.
	for name, count in pairs(entity_module_inventory.get_contents()) do
		local inserted_count = player.insert{name = name, count = count}
		if inserted_count > 0 then
			text_position = show_flying_text(player, surface, text_position, name, inserted_count)
			count = count - inserted_count
		end
		-- For the remaining modules, order robots to remove.
		if robot_removes_modules_after_player then
			for i = 1, count, 1 do
				create_item_for_robot(player_index, surface, position, force, name)
			end
			entity_module_inventory.clear()
		else
			-- Robot will not help. Only remove the inserted modules.
			if inserted_count > 0 then
				entity_module_inventory.remove{name = name, count = inserted_count}
			end
		end
	end
	
	-- Also destroy all existing module requests.
	remove_all_module_requests_on_entity(entity, surface, position, force)
	
	return text_position
end

-- Removes the given module stack from the given entity. The removed modules will be inserted to player first. The exceeded ones may be taken away by robots, according to the given setting.
-- Returns the text position for the next flying text.
-- Caution, make sure the entity has such module of such count before calling this method.
-- @param text_position	Optional.
local function remove_modules_from_entity(entity, entity_module_inventory, player, player_index, surface, position, force, module_stack, robot_removes_modules_after_player, text_position)
	local name = module_stack.name
	local count = module_stack.count
	local text_position = text_position or {x = position.x, y = position.y}
	local inserted_count = player.insert(module_stack)
	if inserted_count > 0 then
		text_position = show_flying_text(player, surface, text_position, name, inserted_count)
		count = count - inserted_count
	end
	-- For the remaining modules, order robots to remove.
	if robot_removes_modules_after_player then
		for i = 1, count, 1 do
			create_item_for_robot(player_index, surface, position, force, name)
		end
		entity_module_inventory.remove(module_stack)
	else
		-- Robot will not help. Only remove the inserted modules.
		if inserted_count > 0 then
			entity_module_inventory.remove{name = name, count = inserted_count}
		end
	end
	return text_position
end

-------------------------------------------------------------------------------------------------

script.on_event(defines.events.on_entity_settings_pasted, function(event)
	local player_index = event.player_index
	local player = game.players[player_index]
	local player_settings = player.mod_settings
	local force = player.force
	
	-- 1. Simple checking.
	
	-- Check is the function enabled.
	if not player_settings["copy-and-paste-modules_enable"].value then
		return
	end
	
	-- Get destination module inventory.
	local dest_entity = event.destination
	local dest_module_inventory = dest_entity.get_module_inventory()
	if not dest_module_inventory then
		if is_debug then game.print("Destination does not have module inventory.") end
		return
	end
	
	-- Is there any slot in the destination? (Assembling machine MK1 has module inventory but no slot.)
	local dest_module_inventory_size = #dest_module_inventory
	if dest_module_inventory_size <= 0 then
		if is_debug then game.print("Destination module inventory has no slot.") end
		return
	end
	
	local complete_override = player_settings["copy-and-paste-modules_complete-override"].value
	local robot_removes_modules_after_player = player_settings["copy-and-paste-modules_robot-removes-modules-after-player"].value	
	
	-- Get source module inventory.
	local source_entity = event.source
	local source_module_inventory = source_entity.get_module_inventory()
	if not source_module_inventory then
		-- Source does not have module inventory.
		if is_debug then game.print("Source does not have module inventory.") end
		-- Complete override: clear the destination.
		if complete_override then
			remove_all_modules_from_entity(dest_entity, dest_module_inventory, player, player_index, dest_entity.surface, dest_entity.position, force, robot_removes_modules_after_player, nil)
		end
		return
	end
	
	---------------------------------------
	
	-- 2. Get the modules to be pasted.
	
	-- Get the modules directly from the module inventory of the source.
	local wanted_modules = source_module_inventory.get_contents()
	
	-- Add the modules from the item request proxies on the source machine.
	local source_position = source_entity.position
	local source_x = source_position.x
	local source_y = source_position.y
	local source_item_request_proxies = source_entity.surface.find_entities_filtered
	{
		area = {{source_x - 0.1, source_y - 0.1}, {source_x + 0.1, source_y + 0.1}},
		name = "item-request-proxy",
		force = force
	}
	for _, proxy in ipairs(source_item_request_proxies) do
		if proxy.valid and proxy.proxy_target == source_entity then
			for name, count in pairs(proxy.item_requests) do
				if game.item_prototypes[name].type == "module" then
					wanted_modules[name] = wanted_modules[name] and wanted_modules[name] + count or count
				end
			end
		end
	end
	
	-- Is there anything to be copied?
	if next(wanted_modules) == nil then
		-- Nothing to be copied.
		if is_debug then game.print("Source does not have any module.") end
		-- Complete override: clear the destination.
		if complete_override then
			remove_all_modules_from_entity(dest_entity, dest_module_inventory, player, player_index, dest_entity.surface, dest_entity.position, force, robot_removes_modules_after_player, nil)
		end
		return
	end
	
	---------------------------------------
	
	-- 3. Count the modules that we can actually insert to destination.
	
	-- Remove those cannot be inserted to the destination (e.g. productivity module).
	--	(This is mainly for Beacon. For Assembling machine, the recipe is also copy-and-pasted, so modules that are in source should also be inserted to destination.)
	-- But first, temporarily remove the first item in the destination inventory, so LuaInventory::can_insert returns false only if the module type cannot be inserted, not because the inventory is full.
	local first_module_name = nil
	if dest_module_inventory[1].valid_for_read then
		first_module_name = dest_module_inventory[1].name
		dest_module_inventory[1].clear()
	end
	
	-- Reduce the requests by the existing modules in the destination.
	-- We will need to know if there are sufficient slots for the modules.
	local wanted_module_count = 0
	
	-- We will need to know the original requested modules that can be inserted, so we can remove those unwanted ones. We only need their names.
	-- Mainly for complete-override.
	local source_module_names = {}
	
	-- Exceeded modules that will be removed for complete-override.
	local exceeded_modules
	
	for name, count in pairs(wanted_modules) do
		if is_debug then game.print("Modules from source: " .. name .. " x" .. count) end
		if dest_module_inventory.can_insert{name = name} then
			-- Can insert module to destination.
			source_module_names[name] = true
			-- How many modules we need and how many already exist?
			count = count - dest_module_inventory.get_item_count(name)
			-- Remember to count the first module that we have temporarily removed.
			if first_module_name and first_module_name == name then
				count = count - 1
			end
			if count <= 0 then
				-- Destination has more modules than we want.
				if is_debug then game.print("    But destination already have enough module: " .. name .. ", count: " .. count) end
				wanted_modules[name] = nil
				-- Complete override: remove the exceeded ones.
				if complete_override and count < 0 then
					-- Don't remove them now, because the first module is not in the entity.
					exceeded_modules = exceeded_modules or {}
					exceeded_modules[name] = -count
				end
			else
				-- Destination has fewer modules that we want.
				if is_debug then game.print("    Destination needs module: " .. name .. " x" .. count) end
				wanted_modules[name] = count
				wanted_module_count = wanted_module_count + count
			end
		else		
			-- Cannot insert module to destination.
			if is_debug then game.print("    Cannot insert module to destination: " .. name) end
			wanted_modules[name] = nil
		end
	end
	-- Insert the first module back.
	if first_module_name then
		dest_module_inventory[1].set_stack{name = first_module_name}
	end
	
	-- Remove the exceeded modules.
	-- Prepare for flying-texts.
	local surface = dest_entity.surface
	local position = dest_entity.position
	local x = position.x
	local y = position.y
	local text_position = {x = x, y = y}
	if exceeded_modules then
		for name, count in pairs(exceeded_modules) do
			text_position = remove_modules_from_entity(dest_entity, dest_module_inventory, player, player_index, surface, position, force, {name = name, count = count}, robot_removes_modules_after_player, text_position)
		end
	end
	
	-- At this point,
	-- wanted_modules = the modules that we actually need to be inserted.
	-- source_module_names = those from the source and we can insert to destination.
	-- Exceeded modules in the destination have been removed.
	-- But those modules that are not in the source but only in the destination have not been handled.
	
	-- Do we actually need to insert items?
	if next(wanted_modules) == nil then
		if is_debug then game.print("No module to request.") end
		-- Handle complete override now.
		if complete_override then
			-- Remove those that are not in the source machine but only in the destination.
			for name, count in pairs(dest_module_inventory.get_contents()) do
				if not source_module_names[name] then
					text_position = remove_modules_from_entity(dest_entity, dest_module_inventory, player, player_index, surface, position, force, {name = name, count = count}, robot_removes_modules_after_player, text_position)
				end
			end
			-- Also remove all module requests.
			remove_all_module_requests_on_entity(dest_entity, surface, position, force)
		end
		return
	end
	
	---------------------------------------
	
	-- 4. Get the number of available slots in the destination.
	
	local available_slots = 0
	
	-- We will also remove the modules we don't want.
	local remove_unwanted_modules = complete_override or player_settings["copy-and-paste-modules_replace-unwanted-modules-for-insufficient-slots"].value
	local remove_slot_indexes = nil
	
	-- How many empty slots are available in the destiantion?
	local available_slots = 0
	if dest_module_inventory.is_empty() then
		available_slots = dest_module_inventory_size
	else
		for i = 1, dest_module_inventory_size, 1 do
			local slot = dest_module_inventory[i]
			if slot.valid_for_read then
				-- Slot is not empty.
				if remove_unwanted_modules and not source_module_names[slot.name] then
					-- It is not a module we want.
					-- Record the index of this slot as a candidate for item removal.
					remove_slot_indexes = remove_slot_indexes or {}
					table_insert(remove_slot_indexes, i)
					if is_debug then game.print("Slot[" .. i .. "] contains the item we don't want. Recorded it for removal.") end
				end
			else
				-- Slot is empty.
				available_slots = available_slots + 1
				if is_debug then game.print("Slot[" .. i .. "] is empty.") end
			end
		end
	end
	if is_debug then game.print("Number of empty slots before item removal: " .. available_slots .. ", and we want " .. wanted_module_count .. " slots.") end
	
	-- Remove unwanted items if not enough empty slots.
	-- Record the modules that have been inserted to player, for creating flying-texts.
	local modules_inserted_to_player
	local enough_slot = wanted_module_count <= available_slots
	if (complete_override or not enough_slot) and remove_slot_indexes then
		for _, slot_index in ipairs(remove_slot_indexes) do
			if is_debug then game.print("Try to clear the slot[" .. slot_index .. "]") end
			local name = dest_module_inventory[slot_index].name
			local is_cleared = false
			-- Insert the module to player's inventory.
			if player.insert{name = name} > 0 then
				-- Inserted to player.
				is_cleared = true
				modules_inserted_to_player = modules_inserted_to_player or {}
				modules_inserted_to_player[name] = modules_inserted_to_player[name] and modules_inserted_to_player[name] + 1 or 1
			else
				-- Order robot to help.
				if robot_removes_modules_after_player then
					if is_debug then game.print("Remove item by robot: " .. name) end
					create_item_for_robot(player_index, surface, position, force, name)
					is_cleared = true
				end
			end
			if is_cleared then
				-- Clear the slot.
				dest_module_inventory[slot_index].clear()
				available_slots = available_slots + 1
				enough_slot = wanted_module_count <= available_slots
				if not complete_override and enough_slot then
					-- We have enough slots now, and complete override is off. Don't need to empty anymore slots.
					break
				end
			end
		end
	end
	if modules_inserted_to_player then
		text_position = show_flying_texts_for_all_items(player, surface, text_position, modules_inserted_to_player)
	end
	
	if is_debug then game.print("Number of empty slots after item removal: " .. available_slots) end
	
	-- Do we have any empty slot now?
	if available_slots <= 0 then
		if is_debug then game.print("No empty slot in the destination.") end
		return
	end
	
	---------------------------------------
	
	-- 5. Reduce exceeded requests.
	
	-- In case the number of empty slots is smaller than the number of requests, reduce the number of requests.
	-- Because we have ensured the total number of empty slots is larger than 0, there should be at least one module can be inserted into destination.
	local temp_available_slots = available_slots
	for name, count in pairs(wanted_modules) do
		if temp_available_slots <= 0 then
			-- No more empty slot for more requests.
			wanted_modules[name] = nil
			if is_debug then game.print("No more empty slot for module: " .. name) end
		else
			count = math_min(count, temp_available_slots)
			wanted_modules[name] = count
			temp_available_slots = temp_available_slots - count
			if is_debug then game.print("Destination can insert module: " .. name .. " x" .. count) end
		end
	end
	
	---------------------------------------
	
	-- What we have done so far:
	-- Ensured there are enough slots for the modules.
	-- The exceeded modules in the destination has been removed, if complete override is enabled.
	
	-- What we will need to do:
	-- Insert the wanted modules.
	-- Update or remove the existing item request proxies on the destination.
	
	---------------------------------------
	
	-- 6. Insert items from player to machine.
	
	-- At the same time, we will deal with the existing item request proxies on the destination.
	local existing_item_request_proxies = surface.find_entities_filtered
	{
		area = {{x - 0.1, y - 0.1}, {x + 0.1, y + 0.1}},
		name = "item-request-proxy",
		force = force
	}
	-- Store all of their item requests.
	local existing_item_requests_by_proxy = {} -- All requests mapped to each proxy.
	local existing_item_request_count_by_name = {} -- All requests for all proxies.
	-- For removing existing item requests, we will first update existing_item_request_count_by_name.
	-- After it is finalized, we will use it to update existing_item_requests_by_proxy, and finally update the proxies.
	for i = #existing_item_request_proxies, 1, -1 do
		local proxy = existing_item_request_proxies[i]
		if proxy.valid and proxy.proxy_target == dest_entity then
			existing_item_requests_by_proxy[proxy] = proxy.item_requests
			for name, count in pairs(proxy.item_requests) do
				existing_item_request_count_by_name[name] = existing_item_request_count_by_name[name] and existing_item_request_count_by_name[name] + count or count
			end
		else
			table_remove(existing_item_request_proxies, i)
		end
	end
	
	-- Insert items.
	for name, count in pairs(wanted_modules) do
		-- Remove items from player's inventory and see how many are actually removed.
		local insert_count = player.remove_item{name = name, count = count}
		if insert_count > 0 then
			-- Insert to the machine.
			-- We have already ensured the destination can accept all of these modules.
			dest_module_inventory.insert{name = name, count = insert_count}
			available_slots = available_slots - insert_count
			-- Reduce or remove item request.
			count = count - insert_count
			if count > 0 then
				wanted_modules[name] = count
			else
				wanted_modules[name] = nil
			end
			-- Reduce or remove item request from the existing proxies.
			-- It is needed only if complete override is disabled.
			-- We will override the proxies if complete override is enabled.
			if not complete_override then
				if existing_item_request_count_by_name[name] then
					local final_count = existing_item_request_count_by_name[name] - insert_count
					if final_count > 0 then
						if is_debug then game.print("Existing item request is reduced from " .. name .. " x" .. existing_item_request_count_by_name[name] .. " to x" .. final_count) end
						existing_item_request_count_by_name[name] = final_count
					else
						if is_debug then game.print("Existing item request of " .. name .. " is fulfilled.") end
						existing_item_request_count_by_name[name] = nil
					end
				end
			end
			-- Show flying-text.
			text_position = show_flying_text(player, surface, text_position, name, -insert_count)
		end
	end
	
	-- Have all wanted modules been inserted?
	if next(wanted_modules) == nil then
		-- Robot delivery is not needed.
		if is_debug then game.print("All modules have been inserted by player. Robot delivery is not needed.") end
		
		-- Finalize the existing item request proxies.
		if complete_override then
			-- Complete override: remove all existing module requests.
			for proxy, _ in pairs(existing_item_requests_by_proxy) do
				remove_all_module_requests_in_item_request_proxy(proxy)
			end
			if is_debug then game.print("Complete override: removed all item request proxies after player insertion.") end
			return
		end
		
		-- Not complete override: update the existing request proxies.
		-- At this point,
		-- existing_item_request_count_by_name = the original requests - the inserted modules.
		update_item_requests_in_proxies(existing_item_requests_by_proxy, existing_item_request_count_by_name, available_slots)
		if is_debug then game.print("Not complete override: updated all item request proxies after player insertion.") end
		return
	end
	
	---------------------------------------
	
	-- 7. Request modules from robots.
	
	-- Can robots insert modules?
	if player_settings["copy-and-paste-modules_robot-inserts-modules-after-player"].value then
		-- Yes.
		-- Iterate all existing item requests to find what requests are missing.
		for proxy, item_requests in pairs(existing_item_requests_by_proxy) do
			for name, _ in pairs(item_requests) do
				if wanted_modules[name] then
					-- We want this module.
					-- The count in item_requests[name] is not what we want.
					-- existing_item_request_count_by_name[name] contains the actual number after player insertion was done.
					local count = existing_item_request_count_by_name[name]
					if count then
						count = wanted_modules[name] - count
					else
						count = wanted_modules[name]
					end
					if count <= 0 then
						-- Too many modules have been requested.
						wanted_modules[name] = nil
						-- Complete override: reduce the reuqests.
						if complete_override then
							item_requests[name] = item_requests[name] + count
						end
					else
						-- Too few modules have been requested.
						wanted_modules[name] = count
						-- Don't change the item request yet. We will add that later.
					end
				else
					-- We don't want this module.
					if complete_override then
						-- Complete override: remove the requests that are not needed anymore.
						if game.item_prototypes[name].type == "module" then
							item_requests[name] = nil
						end
					end
				end
			end
			
			if next(item_requests) == nil then
				-- No more item request for the proxy. Destroy it.
				proxy.destroy()
				existing_item_requests_by_proxy[proxy] = nil
				if is_debug then game.print("No more item request for the proxy. Destroy it.") end
			else
				-- Still have requests for the proxy. Update it.
				proxy.item_requests = item_requests
				if is_debug then game.print("Reduced or kept item requests for the proxy.") end
			end
		end
		
		if next(wanted_modules) ~= nil then
			-- Add requests for the remaining modules.
			local proxy, item_requests = next(existing_item_requests_by_proxy)
			if proxy then
				-- We can reuse the existing proxy.
				-- Add requests to it.
				for name, count in pairs(wanted_modules) do
					item_requests[name] = item_requests[name] and item_requests[name] + count or count
				end
				proxy.item_requests = item_requests
				if is_debug then game.print("Added item requests to the existing proxy.") end
			else
				-- No existing proxy we can use. Create a new one.
				proxy = surface.create_entity
				{
					name = "item-request-proxy",
					position = position,
					force = force,
					target = dest_entity,
					modules = wanted_modules
				}
				-- Raise event.
				raise_event(defines.events.on_built_entity,
				{
					created_entity = proxy,
					player_index = player_index
				})
			end
		end
	else
		-- No, we don't order robots to insert the remaining modules.
		if complete_override then
			-- Complete override: remove all existing module requests.
			for proxy, _ in pairs(existing_item_requests_by_proxy) do
				remove_all_module_requests_in_item_request_proxy(proxy)
			end
			if is_debug then game.print("Complete override: removed all item request proxies without robot insertion.") end
			return
		end
		
		-- Not complete override: update the existing request proxies.
		update_item_requests_in_proxies(existing_item_requests_by_proxy, existing_item_request_count_by_name, available_slots)
		if is_debug then game.print("Not complete override: updated all item request proxies without robot insertion.") end
	end
end)